from __future__ import absolute_import, division, print_function, unicode_literals

import os
import base64
import binascii

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.hmac import HMAC

ALGO_AES_CTR = 'aes-ctr'
ALGO_AES_CTR_LEGACY = 'aes-ctr-legacy'


DEFAULT_KEY_LENGTH = 64

_hash_classes = {
    'SHA224': hashes.SHA224,
    'SHA256': hashes.SHA256,
    'SHA384': hashes.SHA384,
    'SHA512': hashes.SHA512,
}

DEFAULT_DIGEST = 'SHA256'
HASHING_ALGORITHMS = list(_hash_classes.keys())
LEGACY_NONCE = b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01'


def open_aes_ctr(key_service, material):
    """
    Decrypts secrets stored by `seal_aes_ctr`.

    Allows for binary plaintext
    """
    key = key_service.decrypt(material['key'])
    digest_method = material.get('digest', DEFAULT_DIGEST)
    ciphertext = material['contents']
    hmac = material['hmac']
    nonce = material.get('nonce', LEGACY_NONCE)
    return _open_aes_ctr(key, nonce, ciphertext, hmac, digest_method)


def seal_aes_ctr(key_service, secret, digest_method=DEFAULT_DIGEST, key_length=DEFAULT_KEY_LENGTH, binary_type=None):
    """
    Encrypts `secret` using the key service.

    You can decrypt with the companion method `open_aes_ctr`.
    """
    key, encoded_key = key_service.generate_key_data(key_length)
    nonce = os.urandom(16)
    ciphertext, hmac = _seal_aes_ctr(
        secret, key, nonce, digest_method
    )

    # Agh! a mighty break in abstraction
    # DynamoDB wont put `bytes` => `Binary` in python2
    # So we need to wrap it with a special type
    def Binary(value):
        if binary_type:
            return binary_type(value)
        return value
    return {
        'key': Binary(encoded_key),
        'contents': Binary(ciphertext),
        'hmac': Binary(hmac),
        'nonce': Binary(nonce),
        'digest': digest_method,
        'algorithm': ALGO_AES_CTR,
    }


def open_aes_ctr_legacy(key_service, material):
    """
    Decrypts secrets stored by `seal_aes_ctr_legacy`.

    Assumes that the plaintext is str (non-binary).
    """
    key = key_service.decrypt(_from_b64(material['key']))
    digest_method = material.get('digest', DEFAULT_DIGEST)
    ciphertext = _from_b64(material['contents'])
    hmac = _from_hex(material['hmac'])
    return _open_aes_ctr(key, LEGACY_NONCE, ciphertext, hmac, digest_method)


def seal_aes_ctr_legacy(key_service, secret, digest_method=DEFAULT_DIGEST):
    """
    :deprecated: Please use `seal_aes_ctr` instead.

    Encrypts `secret` using the key service.

    You can decrypt with the companion method `open_aes_ctr_legacy`.
    """
    # generate a a 64 byte key.
    # Half will be for data encryption, the other half for HMAC
    key, encoded_key = key_service.generate_key_data(64)
    ciphertext, hmac = _seal_aes_ctr(
        secret, key, LEGACY_NONCE, digest_method,
    )
    return {
        'key': _to_b64(encoded_key),
        'contents': _to_b64(ciphertext),
        'hmac': _to_hex(hmac),
        'digest': digest_method,
    }


def _open_aes_ctr(key, nonce, ciphertext, expected_hmac, digest_method):
    data_key, hmac_key = _halve_key(key)
    hmac = _get_hmac(hmac_key, ciphertext, digest_method)
    # Check the HMAC before we decrypt to verify ciphertext integrity
    if hmac != expected_hmac:
        raise IntegrityError("Computed HMAC on %s does not match stored HMAC")

    decryptor = Cipher(
        algorithms.AES(data_key),
        modes.CTR(nonce),
        backend=default_backend()
    ).decryptor()
    return decryptor.update(ciphertext) + decryptor.finalize()


def _seal_aes_ctr(plaintext, key, nonce, digest_method):
    data_key, hmac_key = _halve_key(key)
    encryptor = Cipher(
        algorithms.AES(data_key),
        modes.CTR(nonce),
        backend=default_backend()
    ).encryptor()

    ciphertext = encryptor.update(plaintext) + encryptor.finalize()
    return ciphertext, _get_hmac(hmac_key, ciphertext, digest_method)


def _get_hmac(key, ciphertext, digest_method):
    hmac = HMAC(
        key,
        get_digest(digest_method),
        backend=default_backend()
    )
    hmac.update(ciphertext)
    return hmac.finalize()


def _halve_key(key):
    half = len(key) // 2
    return key[:half], key[half:]


def get_digest(digest):
    try:
        return _hash_classes[digest]()
    except KeyError:
        raise ValueError("Could not find " + digest + " in cryptography.hazmat.primitives.hashes")


def _to_hex(b):  # type (bytes): -> str
    return binascii.hexlify(b).decode('ascii')


def _from_hex(s):  # type: (str) -> bytes
    return binascii.unhexlify(s)


def _to_b64(b):  # type (bytes): -> str
    b2 = base64.b64encode(b)
    return b2.decode('ascii')


def _from_b64(s):  # type (str): -> bytes
    b = s.encode('ascii')
    return base64.b64decode(b)


class IntegrityError(Exception):

    def __init__(self, value=""):
        self.value = "INTEGRITY ERROR: " + value if value is not "" else \
                     "INTEGRITY ERROR"

    def __str__(self):
        return self.value



from __future__ import absolute_import, division, print_function, unicode_literals

import logging

logger = logging.getLogger(__name__)


def prune_secret(storage_service, secret_name):
    secrets = storage_service.list_one(secret_name)
    if not secrets:
        logger.info('Not found: %s', secret_name)
        return

    max_version = max(
        secret['version']
        for secret in secrets
    )

    for secret in secrets:
        if secret['version'] == max_version:
            continue
        logger.info("Deleting %s -- version %s",
                    secret["name"], secret["version"])
        storage_service.delete_one(
            secret['name'], secret['version']
        )
    logger.info('Pruned %s (current version=%d)', secret_name, max_version)



from __future__ import absolute_import, division, print_function, unicode_literals


import six
from .aes_ctr import open_aes_ctr_legacy, seal_aes_ctr_legacy, ALGO_AES_CTR_LEGACY, \
    open_aes_ctr, seal_aes_ctr, ALGO_AES_CTR
from .aes_gcm import open_aes_gcm, seal_aes_gcm, ALGO_AES_GCM


def seal_secret(key_service, plaintext, algorithm=ALGO_AES_CTR, binary_type=None, **seal_kwargs):
    if isinstance(plaintext, six.text_type):
        plaintext = plaintext.encode('utf-8')

    if not algorithm:
        algorithm = ALGO_AES_CTR

    if algorithm == ALGO_AES_GCM:
        return seal_aes_gcm(
            key_service,
            plaintext,
            binary_type=binary_type,
            **seal_kwargs
        )
    if algorithm == ALGO_AES_CTR:
        return seal_aes_ctr(
            key_service,
            plaintext,
            binary_type=binary_type,
            **seal_kwargs
        )
    if algorithm == ALGO_AES_CTR_LEGACY:
        return seal_aes_ctr_legacy(
            key_service,
            plaintext,
            **seal_kwargs
        )
    raise RuntimeError('Unsupported algo: %s' % algorithm)


def open_secret(key_service, ciphertext):
    algorithm = ciphertext.get('algorithm')

    if algorithm == ALGO_AES_GCM:
        return open_aes_gcm(key_service, ciphertext)

    if algorithm == ALGO_AES_CTR:
        return open_aes_ctr(key_service, ciphertext)

    if not algorithm or algorithm == ALGO_AES_CTR_LEGACY:
        return open_aes_ctr_legacy(key_service, ciphertext)

    raise RuntimeError('Unsupported algo: %s' % algorithm)




from __future__ import absolute_import, division, print_function, unicode_literals


from .delete import delete_secret
from .get import get_secret
from .list import list_secrets
from .prune import prune_secret
from .put import put_secret



from __future__ import absolute_import, division, print_function

import codecs
import fnmatch
import logging
import operator
import os
import sys

import boto3
import click
import pkg_resources
import six

import credsmash.api
from credsmash.crypto import ALGO_AES_CTR
from credsmash.util import set_stream_logger, detect_format, \
    parse_config, read_one, read_many, write_one, write_many

logger = logging.getLogger(__name__)


class Environment(object):
    def __init__(self, storage_service_name, storage_service_config,
                 key_service_name, key_service_config,
                 algorithm, algorithm_options):
        self._storage_service = None
        self.storage_service_name = storage_service_name
        self.storage_service_config = storage_service_config
        self._key_service = None
        self.key_service_name = key_service_name
        self.key_service_config = key_service_config
        self.algorithm = algorithm
        self.algorithm_options = algorithm_options
        self._session = None

    @property
    def session(self):
        if self._session is None:
           self._session = boto3.Session()
        return self._session

    @staticmethod
    def load_entry_point(group, name):
        entry_points = pkg_resources.iter_entry_points(
            group, name
        )
        for entry_point in entry_points:
            return entry_point.load()
        raise RuntimeError('Not found EntryPoint(group={0},name={1})'.format(group, name))

    @property
    def key_service(self):
        if not self._key_service:
            cls = self.load_entry_point('credsmash.key_service', self.key_service_name)
            self._key_service = cls(
                session=self.session,
                **self.key_service_config
            )
            logger.debug('key_service=%r', self._key_service)
        return self._key_service

    @property
    def storage_service(self):
        if not self._storage_service:
            cls = self.load_entry_point('credsmash.storage_service', self.storage_service_name)
            self._storage_service = cls(
                session=self.session,
                **self.storage_service_config
            )
            logger.debug('storage_service=%r', self._storage_service)
        return self._storage_service

    def __repr__(self):
        return 'Environment(storage_service=%r,key_service=%r)' % (self._storage_service, self._key_service)


@click.group()
@click.option('--config', '-c',
              envvar='CREDSMASH_CONFIG',
              type=click.Path(resolve_path=True),
              default='/etc/credsmash.cfg')
@click.option('--table-name', '-t', default=None,
              help="DynamoDB table to use for "
                   "credential storage")
@click.option('--key-id', '-k', default=None,
              help="the KMS key-id of the master key "
                   "to use. See the README for more "
                   "information. Defaults to alias/credsmash")
@click.option('--context', type=(six.text_type, six.text_type), multiple=True,
              help="the KMS encryption context to use."
                   "Only works if --key-id is passed.")
@click.pass_context
def main(ctx, config, table_name, key_id, context=None):
    config_data = {}
    sections = {}
    if os.path.exists(config):
        with codecs.open(config, 'r') as config_fp:
            sections = parse_config(config_fp)
            config_data = sections.get('credsmash', {})

    if key_id:
        # Using --key-id/-k will ignore the configuration file.
        key_service_name = 'kms'
        key_service_config = {
            'key_id': key_id
        }
        if context:
            key_service_config['encryption_context'] = dict(context)
    else:
        if context:
            logger.warning('--context can only be used in conjunction with --key-id')
        key_service_name = config_data.get('key_service', 'kms')
        key_service_config = sections.get('credsmash:key_service:%s' % key_service_name, {})
        if key_service_name == 'kms':
            key_service_config.setdefault(
                'key_id', config_data.get('key_id', 'alias/credsmash')
            )

    if table_name:
        storage_service_name = 'dynamodb'
        storage_service_config = {
            'table_name': table_name
        }
    else:
        storage_service_name = config_data.get('storage_service', 'dynamodb')
        storage_service_config = sections.get('credsmash:storage_service:%s' % storage_service_name, {})
        if storage_service_name == 'dynamodb':
            storage_service_config.setdefault(
                'table_name', config_data.get('table_name', 'secret-store')
            )

    algorithm = config_data.get('algorithm', ALGO_AES_CTR)
    algorithm_options = sections.get('credsmash:%s' % algorithm, {})

    set_stream_logger(
        level=config_data.get('log_level', 'INFO')
    )
    env = Environment(
        storage_service_name,
        storage_service_config,
        key_service_name,
        key_service_config,
        algorithm,
        algorithm_options
    )
    ctx.obj = env


@main.command('list')
@click.argument('pattern', required=False)
@click.pass_context
def cmd_list(ctx, pattern=None):
    """
    List all secrets & their versions.
    """
    secrets = credsmash.api.list_secrets(
        ctx.obj.storage_service
    )
    if pattern:
        matched_names = set(fnmatch.filter((secret['name'] for secret in secrets), pattern))
        secrets = [
            secret
            for secret in secrets
            if secret['name'] in matched_names
        ]

    if not secrets:
        return

    max_len = max(len(secret["name"]) for secret in secrets)
    for cred in sorted(secrets, key=operator.itemgetter("name", "version")):
        click.echo(
            "{cred[name]:{l}} -- version {cred[version]:>}".format(l=max_len, cred=cred)
        )


@main.command('delete')
@click.argument('secret_name')
@click.pass_context
def cmd_delete_one(ctx, secret_name):
    """
    Delete every version of a single secret
    """
    credsmash.api.delete_secret(
        ctx.obj.storage_service, secret_name
    )


@main.command('delete-many')
@click.argument('pattern')
@click.pass_context
def cmd_delete_many(ctx, pattern):
    """
    Delete every version of all matching secrets
    """
    secret_names = {
        x["name"] for x in credsmash.api.list_secrets(ctx.obj.storage_service)
    }
    secret_names = fnmatch.filter(secret_names, pattern)
    for secret_name in secret_names:
        credsmash.api.delete_secret(
            ctx.obj.storage_service, secret_name
        )


@main.command('prune')
@click.argument('secret_name')
@click.pass_context
def cmd_prune_one(ctx, secret_name):
    """
    Delete all but the latest version of a single secret
    """
    credsmash.api.prune_secret(
        ctx.obj.storage_service, secret_name
    )


@main.command('prune-many')
@click.argument('pattern')
@click.pass_context
def cmd_prune_many(ctx, pattern):
    """
    Delete all but the latest version of all matching secrets
    """
    secret_names = {
        x["name"] for x in credsmash.api.list_secrets(ctx.obj.storage_service)
    }
    secret_names = fnmatch.filter(secret_names, pattern)
    for secret_name in secret_names:
        credsmash.api.prune_secret(
            ctx.obj.storage_service, secret_name
        )


@main.command('get')
@click.argument('secret_name')
@click.argument('destination', type=click.File('wb'), required=False, default='-')
@click.option('-f', '--format', 'fmt', default=None)
@click.option('--version', '-v', default=None, type=click.INT)
@click.pass_context
def cmd_get_one(ctx, secret_name, destination, fmt=None, version=None):
    """
    Fetch the latest, or a specific version of a secret
    """
    secret_value = credsmash.api.get_secret(
        ctx.obj.storage_service,
        ctx.obj.key_service,
        secret_name,
        version=version,
    )
    if not fmt:
        fmt = detect_format(destination, default_format='raw')
    write_one(secret_name, secret_value, destination, fmt)


@main.command('get-all')
@click.argument('destination', type=click.File('wb'), required=False, default='-')
@click.option('-f', '--format', 'fmt', default=None)
@click.pass_context
def cmd_get_all(ctx, destination, fmt):
    """
    Fetch the latest version of all secrets
    """
    secret_names = {
        x["name"] for x in credsmash.api.list_secrets(ctx.obj.storage_service)
    }
    secrets = {
        secret_name: credsmash.api.get_secret(
            ctx.obj.storage_service,
            ctx.obj.key_service,
            secret_name,
        )
        for secret_name in secret_names
    }
    if not fmt:
        fmt = detect_format(destination, default_format='json')
    write_many(secrets, destination, fmt)


@main.command('find-one')
@click.argument('pattern')
@click.argument('destination', type=click.File('wb'), required=False, default='-')
@click.option('-f', '--format', 'fmt', default=None)
@click.option('--version', '-v', default=None)
@click.pass_context
def cmd_find_one(ctx, pattern, destination, fmt=None, version=None):
    """
    Find exactly one secret matching <pattern>
    """
    secret_names = {
        x["name"] for x in credsmash.api.list_secrets(ctx.obj.storage_service)
    }
    secret_names = fnmatch.filter(secret_names, pattern)
    if not secret_names:
        raise click.ClickException('No matching secrets found for pattern={0}'.format(pattern))
    if len(secret_names) > 1:
        raise click.ClickException('Too many results ({0}) for pattern={1}'.format(len(secret_names), pattern))

    secret_name = secret_names[0]
    secret_value = credsmash.api.get_secret(
        ctx.obj.storage_service,
        ctx.obj.key_service,
        secret_name,
        version=version,
    )
    if not fmt:
        fmt = detect_format(destination, default_format='raw')
    write_one(secret_name, secret_value, destination, fmt)


@main.command('find-many')
@click.argument('pattern')
@click.argument('destination', type=click.File('wb'), required=False, default='-')
@click.option('-f', '--format', 'fmt', default=None)
@click.pass_context
def cmd_find_many(ctx, pattern, destination, fmt=None):
    """
    Find all secrets matching <pattern>
    """
    secret_names = {
        x["name"] for x in credsmash.api.list_secrets(ctx.obj.storage_service)
    }
    secret_names = fnmatch.filter(secret_names, pattern)
    secrets = {
        secret_name: credsmash.api.get_secret(
            ctx.obj.storage_service,
            ctx.obj.key_service,
            secret_name,
        )
        for secret_name in secret_names
    }
    if not fmt:
        fmt = detect_format(destination, default_format='json')
    write_many(secrets, destination, fmt)


@main.command('put')
@click.argument('secret_name')
@click.argument('source', type=click.File('rb'))
@click.option('-f', '--format', 'fmt', default=None)
@click.option('--version', '-v', default=None, type=click.INT)
@click.option('--compare/--no-compare', default=True,
              help="Compare with the latest value, and skip if unchanged.")
@click.pass_context
def cmd_put_one(ctx, secret_name, source, fmt=None, version=None, compare=True):
    """
    Store a secret
    """
    if not fmt:
        fmt = detect_format(source, default_format='raw')
    secret_value = read_one(secret_name, source, fmt)

    stored_version = credsmash.api.put_secret(
        ctx.obj.storage_service,
        ctx.obj.key_service,
        secret_name,
        secret_value,
        version=version,
        compare=compare,
        algorithm=ctx.obj.algorithm,
        **ctx.obj.algorithm_options
    )
    logger.info(
        'Stored {0} @ version {1}'.format(secret_name, stored_version)
    )


@main.command('put-many')
@click.argument('source', type=click.File('rb'))
@click.option('-f', '--format', 'fmt', default=None)
@click.option('--compare/--no-compare', default=True,
              help="Compare with the latest value, and skip if unchanged.")
@click.pass_context
def cmd_put_many(ctx, source, fmt, compare=True):
    """
    Store many secrets
    """
    if not fmt:
        fmt = detect_format(source, default_format='json')
    secrets = read_many(source, fmt)

    for secret_name, secret_value in secrets.items():
        stored_version = credsmash.api.put_secret(
            ctx.obj.storage_service,
            ctx.obj.key_service,
            secret_name,
            secret_value,
            version=None,
            compare=compare,
            algorithm=ctx.obj.algorithm,
            **ctx.obj.algorithm_options
        )
        logger.info('Stored {0} @ version {1}'.format(secret_name, stored_version))
    logger.debug('Stored {0} secrets'.format(len(secrets)))


# Load any extra CLI's
for ep in pkg_resources.iter_entry_points('credsmash.cli'):
    try:
        ep.load()
    except ImportError:
        pass
